#!/usr/bin/env python3
#
# This is a simulation for clock synchronization in a WSN
#
#

import random
import time

topo={}

topo[1] = [2, 3, 4]
topo[2] = [1, 3, 4, 6, 7, 8]
topo[3] = [1, 2, 5, 8]
topo[4] = [1, 2, 6]
topo[5] = [3]
topo[6] = [2, 4, 7]
topo[7] = [2, 6, 8]
topo[8] = [2, 3, 7]

#topo = {}
#topo[1] = [2]
#topo[2] = [1]

n = len(topo)

minsleep = 20.0
minactive = 2.0

# Sleep time: number of seconds of sleep
# stime[i] = list of sleep times known to node i
stime = [[minsleep for i in range(n)] for j in range(n)]

# Current time. This increments with the "minactive" and sleep times
ctime = [0.0 for i in range(n)]

# Inbox where nodes receive the sleep times of other nodes
inbox = [[0.0 for i in range(n)] for j in range(n)]


def display(tm):
    for line in tm:
        for v in line:
            print("{:.3} ".format(v), end='')
        print("")
    print("")

# Send a broadcast announcing your next sleep time
def broadcast(src, sleep):
    for v in topo[src + 1]:
        inbox[v - 1][src] = sleep

# Introduce randomness (desynchronization) in the sleep times
def desync(tm):
    for i in range(len(tm)):
        x = random.random() * 0.5 - 0.25
        for j in range(len(tm[i])):
            tm[i][j] = tm[i][j] + x

# Determine the maximum difference between 2 elements in "v"
def maxdif(v):
    a, b = 0, 0
    max = 0
    for i in range(len(v)):
        for j in range(i, len(v)):
            if abs(v[i] - v[j]) > max:
                a, b = i, j
                max = abs(v[i] - v[j])
    return a, b, max

random.seed()

#desync(stime)
display(stime)

max_all_time_diff = 0
avg_diff = 0
avg_count = 0

for step in range(0, 100):

    # Artificial desynchronization
    if step % 20 == 0:
        desync(stime)

    # Broadcast the sleep time of each node
    for i in range(0, n):
        broadcast(i, stime[i][i])
        #display(inbox)

        # This is where the node does it's work

        # Active period passes
        ctime[i] = ctime[i] + minactive

    
    a, b, m = maxdif(ctime)
    max_all_time_diff = max(max_all_time_diff, m)
    avg_diff = avg_diff + m
    avg_count = avg_count + 1

    print(m)

    # Make a copy of the current time vector. It will be modified by each node
    # Normally, this will also be sent with a broadcast. Here this is a global
    # vector to make it simpler.
    ntime = ctime[:]

    print("node\ttime\tsleep\tnext_sleep")
    # Calculate the next sleep time before going to sleep
    for i in range(0, n):
        next_time = stime[i][i]
        count = 1

        for v in topo[i + 1]:
            next_time = next_time + inbox[i][v - 1] + ctime[v - 1] - ctime[i]
            #next_time = next_time * (inbox[i][v - 1] + ctime[v - 1] - ctime[i])
            count = count + 1

        next_time = next_time / count
        #next_time = pow(next_time, 1 / count)
        next_time = max(minsleep, next_time)

        print("{}\t{:.4}\t{:.4}\t{:.4}".format(i, ctime[i], stime[i][i], next_time))

        # Sleep period passes
        ntime[i] = ctime[i] + stime[i][i]
        
        # Make the correction to sleep time
        stime[i][i] = next_time

    ctime = ntime[:]

    time.sleep(0.1)
    print("")

print("Max all time diff: ", max_all_time_diff)
avg_diff = avg_diff / avg_count
print("Avg diff: ", avg_diff)
